#include <string.h>
#include <omnetpp.h>
#include <netinet/ip.h>
#include <netinet/udp.h>
#include "InterfaceTable.h"
#include "InterfaceTableAccess.h"
#include "IPDatagram_m.h"
#include "UDPPacket.h"

#include "TunOutDevice.h"


Define_Module(TunOutDevice);

void TunOutDevice::initialize(int stage)
{
    error("The TunOut underlay doesn't work in this version (work in progress)!");

    if (stage==3) {
        // update display string when addresses have been autoconfigured etc.
        updateDisplayString();
        return;
    }

    // all initialization is done in the first stage
    if (stage!=0)
        return;

    interfaceEntry = NULL;
    packetNotification = new cMessage("packetNotification");
    mtu = par("mtu");

    scheduler = check_and_cast<cTunOutScheduler *>(simulation.scheduler());
    scheduler->setInterfaceModule(this, packetNotification, &packetBuffer, mtu);

    parser = check_and_cast<PacketParser*>(parentModule()->submodule("packetParser"));
    numSent = numRcvdOK = numRcvError = numSendError = 0;
    WATCH(numSent);
    WATCH(numRcvdOK);
    WATCH(numRcvError);
    WATCH(numSendError);

#ifdef _MAX_SPEED

    gateIndexNetwOut = gate("netwOut")->id();
#endif

    // register our interface entry in RoutingTable
    interfaceEntry = registerInterface();

}

InterfaceEntry *TunOutDevice::registerInterface()
{
    InterfaceEntry *e = new InterfaceEntry();

    // interface name: our module name without special characters ([])
    char *interfaceName = new char[strlen(parentModule()->fullName())+1];
    char *d=interfaceName;
    for (const char *s=parentModule()->fullName(); *s; s++)
        if (isalnum(*s))
            *d++ = *s;
    *d = '\0';

    e->setName(interfaceName);
    delete [] interfaceName;

    // port: index of gate where our "netwIn" is connected (in IP)
    int outputPort = /*parentModule()->*/gate("netwIn")->fromGate()->index();
    e->setNodeOutputGateId(outputPort);

    // generate a link-layer address to be used as interface token for IPv6
    InterfaceToken token(0, simulation.getUniqueNumber(), 64);
    e->setInterfaceToken(token);

    // MTU: typical values are 576 (Internet de facto), 1500 (Ethernet-friendly),
    // 4000 (on some point-to-point links), 4470 (Cisco routers default, FDDI compatible)
    e->setMtu(mtu);

    // capabilities
    e->setMulticast(true);
    e->setPointToPoint(true);

    // add
    InterfaceTable *ift = InterfaceTableAccess().get();
    ift->addInterface(e, NULL);

    return e;
}


void TunOutDevice::handleMessage(cMessage *msg)
{
    // Packet from the real world...
    if (msg==packetNotification) {
        EV << "TunOutDevice: Message from outside. Queue length = " << packetBuffer.size() << ".\n";
        while( packetBuffer.size() > 0 ) {
            // get packet from buffer and parse it
            char* buf = *(packetBuffer.begin());
            packetBuffer.pop_front();
            cMessage *parsedPacket = decapsulate(buf);
            if (parsedPacket) {
                numRcvdOK++;
#ifndef _MAX_SPEED

                send(parsedPacket,"netwOut");
#else

                send(parsedPacket, gateIndexNetwOut);
#endif

            } else {
                numRcvError++;
            }

        }
    } else // arrived on gate "netwIn"
    {
        // Packet from inside, send to real word
#ifndef _MAX_SPEED
        EV << "Received " << msg << " for transmission\n"; // ib: speed hack
#endif

        transmitToTun(msg);
    }

    if (ev.isGUI())
        updateDisplayString();

}

void TunOutDevice::transmitToTun(cMessage *msg)
{
    unsigned int length;
    char* buf = encapsulate(msg, &length);
    if( buf ) {
        numSent++;
        scheduler->sendBytes( buf, length );
    } else {
        numSendError++;
    }
    delete buf;
}

char* TunOutDevice::encapsulate(cMessage *msg, unsigned int* length)
{
    struct udppseudohdr {
        uint32_t saddr;
        uint32_t daddr;
        uint8_t zero;
        uint8_t protocol;
        uint16_t lenght;
    }
    * pseudohdr;

    unsigned int payloadlen;
    static unsigned int iplen = 20; // we don't generate IP options
    static unsigned int udplen = 8;

    IPDatagram* IP = check_and_cast<IPDatagram*>(msg);
    // FIXME: Check for ICMP-Messages
    UDPPacket* UDP = check_and_cast<UDPPacket*>(IP->decapsulate());
    cMessage* payloadMsg = UDP->decapsulate();

    // parse payload
    char* payload = parser->encapsulatePayload(payloadMsg, &payloadlen);
    if (!payload )
        return NULL;

    *length = payloadlen + iplen + udplen;
    if( *length > mtu ) {
        EV << "TunOutDevice::encapsulate: Error: Packet too big! Size = " << *length << " MTU = " << mtu << "\n";
        return NULL;
    }

    char* buf = new char[*length];

    // We use the buffer to build an ip packet.
    // To minimise unnecessary copying, we start with the payload
    // and write it to the end of the buffer
    memcpy( (buf + iplen + udplen), payload, payloadlen);

    // write udp header in front of the payload
    udphdr* udp_buf = (udphdr*) (buf + iplen);
    udp_buf->source = htons(UDP->sourcePort());
    udp_buf->dest = htons(UDP->destinationPort());
    udp_buf->len = htons(udplen + payloadlen);
    udp_buf->check = 0;

    // Write udp pseudoheader in from of udp header
    // this will be overwritten by ip header
    pseudohdr = (udppseudohdr*) (buf + iplen - sizeof(struct udppseudohdr));
    uint32_t saddr = htonl(IP->srcAddress().getInt());
    uint32_t daddr = htonl(IP->destAddress().getInt());
    pseudohdr->saddr = saddr;
    pseudohdr->daddr = daddr;
    pseudohdr->zero = 0;
    pseudohdr->protocol = IPPROTO_UDP;
    pseudohdr->lenght = udp_buf->len;

    // compute UDP checksum
    udp_buf->check = cksum((uint16_t*) pseudohdr, sizeof(struct udppseudohdr) + udplen + payloadlen);

    // write ip header to begin of buffer
    iphdr* ip_buf = (iphdr*) buf;
    ip_buf->version = 4; // IPv4
    ip_buf->ihl = iplen / 4;
    ip_buf->tos = IP->diffServCodePoint();
    ip_buf->tot_len = htons(*length);
    ip_buf->id = htons(IP->identification());
    ip_buf->frag_off = htons(IP_DF); // DF, no fragments
    ip_buf->ttl = IP->timeToLive();
    ip_buf->protocol = IPPROTO_UDP;
    ip_buf->saddr = saddr;
    ip_buf->daddr = daddr;
    ip_buf->check = 0;
    ip_buf->check = cksum((uint16_t*) ip_buf, iplen);

    delete IP;
    delete UDP;
    delete payloadMsg;
    delete payload;

    return buf;
}

cMessage* TunOutDevice::decapsulate(char* buf)
{
    // Message starts with IP header
    iphdr* ip_buf = (iphdr*) buf;
    udphdr* udp_buf;
    IPDatagram* IP = new IPDatagram;
    UDPPacket* UDP = new UDPPacket;
    cMessage* payload = 0;
    unsigned int payloadLen, datagramlen;
    unsigned int packetlen = ntohs(ip_buf->tot_len);

    // Parsing of IP header, sanity checks
    if (  packetlen > mtu ) {
        ev << "TunOutDevice: Dropping bogus packet, length = " << packetlen << " but mtu = " << mtu <<".\n";
        goto parse_error;
    }
    if ( ip_buf->version != 4 ) {
        ev << "TunOutDevice: Dropping Packet: Packet is not IPv4.\n";
        goto parse_error;
    }
    if ( ntohs(ip_buf->frag_off) & 0xBFFF ) { // mask DF bit
        ev << "TunOutDevice: Dropping Packet: Can't handle fragmented packets.\n";
        goto parse_error;
    }
    if ( ip_buf->protocol != IPPROTO_UDP ) { // FIXME: allow ICMP packets
        ev << "TunOutDevice: Dropping Packet: Packet is not UDP.\n";
        goto parse_error;
    }
    IP->setSrcAddress( IPAddress( ntohl(ip_buf->saddr) ));
    IP->setDestAddress( IPAddress( ntohl(ip_buf->daddr) ));
    IP->setTransportProtocol( ip_buf->protocol );
    IP->setTimeToLive( ip_buf->ttl );
    IP->setIdentification( ntohs(ip_buf->id) );
    IP->setMoreFragments( false );
    IP->setDontFragment( true );
    IP->setFragmentOffset( 0 );
    IP->setDiffServCodePoint( ip_buf->tos );
    IP->setLength( ip_buf->ihl*32 );
    // FIXME: check IP and UDP checksum...

    // Parse UDP header, sanity checks
    udp_buf = (udphdr*)( ((uint32_t *)ip_buf) + ip_buf->ihl );
    datagramlen = ntohs(udp_buf->len);
    if ( (datagramlen != packetlen - ip_buf->ihl*4) ) {
        ev << "TunOutDevice: Dropping Packet: Bogus UDP datagram length: len = " << datagramlen << " packetlen = " << packetlen << " ihl*4 " << ip_buf->ihl*4 << ".\n";
        goto parse_error;
    }
    UDP->setSourcePort( ntohs( udp_buf->source ));
    UDP->setDestinationPort( ntohs( udp_buf->dest ));
    UDP->setByteLength( sizeof( struct udphdr ) );

    // parse payload
    payloadLen = datagramlen - sizeof( struct udphdr );
    payload = parser->decapsulatePayload( ((char*) udp_buf) + sizeof( struct udphdr ), payloadLen );
    if (!payload) {
        ev << "TunOutDevice: Parsing of Payload failed, dropping packet.\n";
        goto parse_error;
    }
    // encapsulate messages
    UDP->encapsulate( payload );
    IP->encapsulate( UDP );

    delete buf;
    return IP;

    // In case the parsing of the packet failed, free allocated memory
parse_error:
    delete buf;
    delete IP;
    delete UDP;
    return NULL;
}

void TunOutDevice::updateDisplayString()
{
    char buf[80];
    if (ev.disabled()) {
        // speed up things
        displayString().setTagArg("t",0,"");
    }
    sprintf(buf, "rcv:%ld snt:%ld", numRcvdOK, numSent);

    if (numRcvError>0)
        sprintf(buf+strlen(buf), "\nerrin:%ld errout:%ld", numRcvError, numSendError);

    displayString().setTagArg("t",0,buf);
}

