#include "ctunoutscheduler.h"

Register_Class(cTunOutScheduler);


inline std::ostream& operator<<(std::ostream& ev, const timeval& tv)
{
    return ev << (unsigned long)tv.tv_sec << "s" << tv.tv_usec << "us";
}

//---

cTunOutScheduler::cTunOutScheduler() : cScheduler()
{
    tun_fd = -1;
}

cTunOutScheduler::~cTunOutScheduler()
{
    delete dev;
}

void cTunOutScheduler::startRun()
{
    if (initsocketlibonce()!=0)
        throw new cRuntimeError("cTunOutScheduler: Cannot initialize socket library");

    gettimeofday(&baseTime, NULL);

    module = NULL;
    notificationMsg = NULL;

    dev = new char[IFNAMSIZ];
    strcpy(dev, "tun%d");
    tun_alloc();
}

void cTunOutScheduler::tun_alloc()
{
    // see /usr/src/linux/Documentation/network/tuntap.txt
    struct ifreq ifr;
    int err;

    if (tun_fd > 0)
        opp_error("Already bound to TUN device!");
    if( (tun_fd = open("/dev/net/tun", O_RDWR)) < 0 ) {
        opp_error("Error opening tun device");
    } else {
        ev << "cTunOutScheduler: Successfully opened TUN device.\n";
    }

    memset(&ifr, 0, sizeof(ifr));

    /* Flags: IFF_TUN   - TUN device (no Ethernet headers)
     *        IFF_TAP   - TAP device  
     * 
     *        IFF_NO_PI - Do not provide packet information  
     */
    ifr.ifr_flags = IFF_TUN | IFF_NO_PI;
    if( *dev ) {
        strncpy(ifr.ifr_name, dev, IFNAMSIZ);
    }

    if((err = ioctl(tun_fd, TUNSETIFF, (void *) &ifr)) < 0 ) {
        close(tun_fd);
        opp_error("Error ioctl tun device");
    } else {
        strncpy(dev, ifr.ifr_name, IFNAMSIZ);
        ev << "cTunOutScheduler: Bound to device " << dev << "\n";
    }
}

void cTunOutScheduler::endRun()
{}

void cTunOutScheduler::executionResumed()
{
    gettimeofday(&baseTime, NULL);
    baseTime = timeval_substract(baseTime, sim->simTime());
}

void cTunOutScheduler::setInterfaceModule(cModule *mod, cMessage *notifMsg,  PacketBuffer* buffer, int mtu)
{
    if (module)
        throw new cRuntimeError("cTunOutScheduler: setInterfaceModule() already called");
    if (!mod || !notifMsg || !buffer)
        throw new cRuntimeError("cTunOutScheduler: setInterfaceModule(): arguments must be non-NULL");

    module = mod;
    notificationMsg = notifMsg;
    packetBuffer = buffer;
    buffersize = mtu;
}

bool cTunOutScheduler::receiveWithTimeout(long usec)
{
    // prepare sets for select()
    fd_set readFD;
    FD_ZERO(&readFD);

    FD_SET(tun_fd, &readFD);

    timeval timeout;
    timeout.tv_sec = 0;
    timeout.tv_usec = usec;

    if (select(FD_SETSIZE, &readFD, NULL, NULL, &timeout) > 0) {
        // Incoming data
        char* buf = new char[buffersize];
        int nBytes = read(tun_fd, buf, buffersize);
        if (nBytes < 0) {
            opp_error("Read from TUN device returned an error");
        } else if (nBytes == 0) {
            opp_error("TUN device closed...");
        } else {
            // write data to buffer
            ev << "cTunOutScheduler: received " << nBytes << " bytes\n";
            packetBuffer->push_back(buf);

            // schedule notificationMsg for the interface module
            timeval curTime;
            gettimeofday(&curTime, NULL);
            curTime = timeval_substract(curTime, baseTime);
            simtime_t t = curTime.tv_sec + curTime.tv_usec*1e-6;
            // TBD assert that it's somehow not smaller than previous event's time
            notificationMsg->setArrival(module,-1,t);
            simulation.msgQueue.insert(notificationMsg);
            return true;
        }
    }
    return false;
}

int cTunOutScheduler::receiveUntil(const timeval& targetTime)
{
    // if there's more than 200ms to wait, wait in 100ms chunks
    // in order to keep UI responsiveness by invoking ev.idle()
    timeval curTime;
    gettimeofday(&curTime, NULL);
    while (targetTime.tv_sec-curTime.tv_sec >=2 ||
            timeval_diff_usec(targetTime, curTime) >= 200000) {
        if (receiveWithTimeout(100000)) // 100ms
            return 1;
        if (ev.idle())
            return -1;
        gettimeofday(&curTime, NULL);
    }

    // difference is now at most 100ms, do it at once
    long usec = timeval_diff_usec(targetTime, curTime);
    if (usec>0)
        if (receiveWithTimeout(usec))
            return 1;
    return 0;
}

cMessage *cTunOutScheduler::getNextEvent()
{
    // assert that we've been configured
    if (!module)
        throw new cRuntimeError("cTunOutScheduler: setInterfaceModule() not called: it must be called from a module's initialize() function");

    // calculate target time
    timeval targetTime;
    cMessage *msg = sim->msgQueue.peekFirst();
    if (!msg) {
        // if there are no events, wait until something comes from outside
        // TBD: obey simtimelimit, cpu-time-limit
        targetTime.tv_sec = LONG_MAX;
        targetTime.tv_usec = 0;
    } else {
        // use time of next event
        simtime_t eventSimtime = msg->arrivalTime();
        targetTime = timeval_add(baseTime, eventSimtime);
    }

    // if needed, wait until that time arrives
    timeval curTime;
    gettimeofday(&curTime, NULL);
    if (timeval_greater(targetTime, curTime)) {
        int status = receiveUntil(targetTime);
        if (status == -1)
            return NULL; // interrupted by user
        if (status == 1)
            msg = sim->msgQueue.peekFirst(); // received something
    } else {
        // we're behind -- customized versions of this class may
        // alert if we're too much behind, whatever that means
    }

    // ok, return the message
    return msg;
}

void cTunOutScheduler::sendBytes(const char *buf, size_t numBytes)
{
    if (!buf) {
        ev << "Error sending packet: buf = NULL!\n";
        return;
    }
    if( numBytes > buffersize ) {
        ev << "trying to send oversizd packet: size " << numBytes << " mtu " << buffersize << "\n";
        opp_error("Can't send packet: too large"); //FIXME: Throw exception instead
    }
    write(tun_fd, buf, numBytes);
    // TBD check for errors
}

