#ifndef __CTUNOUTSCHEDULER_H__
#define __CTUNOUTSCHEDULER_H__

#include "platdep/time.h"
#include "platdep/sockets.h"
// Note: this only works in linux...
#include <linux/if.h>
#include <linux/if_tun.h>
#include <sys/ioctl.h>
//#include <linux/fcntl.h> //for O_RDWR
#include <omnetpp.h>
#include <list>

/** This class implements a event scheduler for omnet
 *  It makes the simulation run in realtime (i.e. 1 simsec == 1 sec)
 *  It is capable of reading/sending data to linux' tun-device
 *  Preparing or interpreting the data is done by other classes, however
 **/
class cTunOutScheduler : public cScheduler
{
public:
    typedef std::list<char*> PacketBuffer;

protected:
    // config
    int tun_fd;
    size_t buffersize;
    char* dev;

    cModule *module;
    cMessage *notificationMsg;
    PacketBuffer *packetBuffer;

    // state
    timeval baseTime;

    /** Allocates the tun device
     **/
    virtual void tun_alloc();

    /**
     * Waits for incoming data on the tun device
     *
     * \param usec Timeout after which to quit waiting (in µsec)
     * \return true if data was read, false otherwise
     **/
    virtual bool receiveWithTimeout(long usec);

    /**
     * Tries to read data until the given time is up
     *
     * \param targetTime stop waiting after this time is up
     * \return 1 if data is read, -1 if there is an error, 0 if no data is read
     **/
    virtual int receiveUntil(const timeval& targetTime);

public:
    /**
     * Constructor.
     */
    cTunOutScheduler();

    /**
     * Destructor.
     */
    virtual ~cTunOutScheduler();

    /**
     * Called at the beginning of a simulation run.
     */
    virtual void startRun();

    /**
     * Called at the end of a simulation run.
     */
    virtual void endRun();

    /**
     * Recalculates "base time" from current wall clock time.
     */
    virtual void executionResumed();

    /**
     * To be called from the module which wishes to receive data from the
     * tun device. The method must be called from the module's initialize()
     * function.
     *
     * \param module Pointer to the module that whishes to receive the data
     * \param notificationMsg A pointer to a message that will be sheduled if there is data to read
     * \param buffer A pointer to the buffer the data will be written into
     * \param mtu Max allowed packet size
     */
    virtual void setInterfaceModule(cModule *module, cMessage *notificationMsg, PacketBuffer* buffer, int mtu);

    /**
     * Scheduler function -- it comes from cScheduler interface.
     */
    virtual cMessage *getNextEvent();

    /**
     * Send data to tun device
     *
     * \param buf A pointer to the data to be send
     * \param numBytes the length of the data
     */
    virtual void sendBytes(const char *buf, size_t numBytes);
};

#endif

